/* Process model C form file: HostRsvpDeamon.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "HostRsvpDeamon.pr.h"
FSM_EXT_DECS



/* Header block */
#include<stdio.h>
#include<errno.h>

#define ARRIVAL op_intrpt_type()==OPC_INTRPT_STRM

#define IP_STREAM 				0
#define APPL_STREAM				1

#define InLIST_SIZE     		10
#define TCP_PROTOCOL_ID			12
#define UDP_PROTOCOL_ID			20
#define RSVP_PROTOCOL_ID 		46

#define REFRESH_PERIOD     		300
#define Path_Msg_Refresh_timeout 4000

#define RSVP_CURRENT_VER		0x01

#define PATH_MESSAGE			01
#define RESV_MESSAGE			02
#define PATH_ERROR				03
#define RESV_ERROR				04
#define PATH_TEAR				05
#define RESV_TEAR				06
#define RESV_CONF				07
#define SESSION_MSG				100
#define SENDER_MSG				101 
#define RESERVE_MSG				102
#define RELEASE_MSG				103

#define MAXNOOFSESSIONS			20
#define GwyAddr					5

#define ON						1
#define OFF						0
#define VL						if(OFF)
#define HRVL					if(OFF)
#define NOV						if(OFF)

#define UNDEFINED               -1
#define DEFAULT_RESERVE_TTL     7
#define MCAST_ADDRESS           500

#ifndef _DEBUG_FUNCS_H
#define _DEBUG_FUNCS_H
 
 
#define PrintErr(p)  \
            printf p ; printf(" : FileName = %s : LineNo = %d :\n",  __FILE__, __LINE__);  printf("ErrNo = %d
: ErrStr = %s :\n\n",  errno, strerror(errno));

#define MALLOC(sz)      calloc(1,sz)
#define Open(x,y,z) \
        open(x,y,z); \
        if( errno ) \
        { PrintErr(("Cannot open file %s ",x)); }
 
#define OpenE(x,y,z) \
        open(x,y,z); \
        if( errno ) \
        { PrintErr( ("Cannot open file %s ",x) ); exit(errno); }
 
#define Malloc(x) \
        malloc(x); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
 
#define MallocE(x) \
        malloc(x); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Calloc(x,y) \
        calloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
 
#define CallocE(x,y) \
        calloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Realloc(x,y) \
        realloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
#define ReallocE(x,y) \
        realloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Read(x,y,z) \
        read(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); }
 
#define ReadE(x,y,z) \
        read(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); exit(errno); }
 
#define Write(x,y,z) \
        write(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); }
 
#define WriteE(x,y,z) \
        write(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); exit(errno); }
 
#define Fopen(x,y) \
        fopen(x,y); \
        if( errno ) \
        { PrintErr(("Cannot open file %s", x)); }
 
#define FopenE(x,y) \
        fopen(x,y); \
        if( errno ) \
        { PrintErr(("Cannot open file %s", x)); exit(errno); }
 
#define Fread(x,y,z,s) \
        fread(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); }
 
#define FreadE(x,y,z,s) \
        fread(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); exit(errno); }
 
#define Fwrite(x,y,z,s) \
        fwrite(x,y,z,s); \
        if( errno ) \
		 { PrintErr(("Cannot read file")); }
 
#define FwriteE(x,y,z,s) \
        fwrite(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); exit(errno); }
 
#define Memcpy(x,y,z) op_prg_mem_copy( y, x, z );
 
#endif

/*
** IP Address structure
*/
typedef struct {
	int NetNum;
	int NodeNum;
} IPAddress;

struct GenList {
	struct GenList *next, *curr;
	char *ele;
};

typedef	struct GenList GenList;

/* 
** Session Information
*/
typedef struct {
	IPAddress DestAddress;
	int ProtocolId;
	int DestPort;
} SessInfo;

/*
** Common Header
*/
typedef struct {
	int Version;
	int Msg_Type;
	int Send_TTL;
	double RSVP_Checksum;
	double RSVP_Length;
} CommonHdr;

/*
** Session object 
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress DestAddress;
	int Protocol_Id;
	int Flags;
	int DestPort;
} SessionObj;

/*
** RSVP Hop Object
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress Next_Hop_Address;
	IPAddress Logical_Interface_Handle;
} RsvpHopObj;

/*
** Time values object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double Refresh_Period;
} TimeValObj;

/*
** Sender template
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress SrcAddress;
	int SrcPort;
} SenderTemplateObj;

/*
** Sender T-Spec object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double req_bandwidth;
	double link_delay;
} SenderTspecObj;

/*
** FlowSpec Object
*/
typedef SenderTspecObj FlowSpecObj;

/*
** Filter Spec
*/
typedef SenderTemplateObj FilterSpecObj;

/*
** Style 
*/
typedef struct {
	int Class_Num;
	double Length;
	int Resv_Style;
} StyleObj;

/*
** Scope object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress SrcAddress;
} ScopeObj;

/*
** Reservation Confirm object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Receiver_Address;
} ResvConfirmObj;

/*
** Error Spec object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Error_Node_Address;
	int Flags;
	int Error_code;
} ErrorSpecObj;



/* 
** Sender Descriptor structure 
*/
typedef struct {

	/* SENDER_TEMPLATE structure */
  SenderTemplateObj Sender_Template_Object; 

	/* SENDER_TSPEC structure */
  SenderTspecObj Sender_Tspec_Object;     

}SenderDescriptor;

/* 
** Flow descriptor list structure 
*/
typedef struct {

	/* FLOWSPEC structure */
  FlowSpecObj Flowspec_Object;         

  /* this was added when it was known that FF style needs to be considered */
	/* FILTER_SPEC structure */
  FilterSpecObj Filter_Spec_Object;      

}FlowDescriptorList;

/* 
** PATH message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header;

	/* SESSION object structure */
  SessionObj Session_Object;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor;

}PathMessage;

/* 
** RESV message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object       ;

	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list[InLIST_SIZE];

}ResvMessage;

/* 
** PathTear message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathTearMessage;

/* 
** ResvTear message structure 
*/
typedef struct {
	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvTearMessage;

/* 
** PathErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathErrMessage;

/* 
** ResvErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object;

	/* STYLE Object structure */
  StyleObj Style_Object;

	/* Flow descriptor list structure */
  /*  ErrorFlowDescriptor error_flow_descriptor;  */

} ResvErrMessage;

/* 
** ResvConf message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;


	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvConfMessage;
	
typedef struct {
	int isfull;
	IPAddress DestAddress;
	int ProtocolId;
	int DestPort;
	void (*UpcallProc)();
} RsvpSession;

typedef struct {
	int SessionID;
	IPAddress SrcAddress;
	int SrcPort;
	SenderTemplateObj Sender_Template_Object;
	SenderTspecObj Sender_Tspec_Object;
	int TTL;
} RsvpSender;

typedef struct{
	int SessionID;
	FlowDescriptorList flow_descriptor_list[InLIST_SIZE];
} RsvpReserve;

/**
*** Function Prototypes
**/

int Session( IPAddress da, int pid, int dp, void (*Up)() );
void HostInitialize( void );
void SendPathMsg( PathMessage pmEle, IPAddress sa );
void MakeSessionCall( Packet* Pkptr );
void MakeSenderCall( Packet* Pkptr );
void MakeReserveCall( Packet* Pkptr );
void MakeReleaseCall( Packet* Pkptr );



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	int                     		sv_RecDataPacket;
	int                     		sv_RsvpUpCalls;
	int                     		sv_SessionId;
	int                     		sv_GrpId[InLIST_SIZE];
	int                     		sv_GrpNo;
	int                     		sv_LinkNo;
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	int                     		sv_InStrm;
	int                     		sv_data_length;
	Packet*                 		sv_RecvPkptr;
	int                     		sv_NoOfLinks;
	IPAddress               		sv_HostIPAddress;
	IPAddress               		sv_LocalAddress[InLIST_SIZE];
	RsvpSession             		sv_rsvpSession[MAXNOOFSESSIONS];
	RsvpSender              		sv_rsvpSender[MAXNOOFSESSIONS];
	RsvpReserve             		sv_rsvpReserve[MAXNOOFSESSIONS];
	struct GenList          		sv_PSBList;
	struct GenList          		sv_RSBList;
	struct GenList          		sv_TCSBList;
	} HostRsvpDeamon_state;

#define pr_state_ptr            		((HostRsvpDeamon_state*) SimI_Mod_State_Ptr)
#define RecDataPacket           		pr_state_ptr->sv_RecDataPacket
#define RsvpUpCalls             		pr_state_ptr->sv_RsvpUpCalls
#define SessionId               		pr_state_ptr->sv_SessionId
#define GrpId                   		pr_state_ptr->sv_GrpId
#define GrpNo                   		pr_state_ptr->sv_GrpNo
#define LinkNo                  		pr_state_ptr->sv_LinkNo
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define InStrm                  		pr_state_ptr->sv_InStrm
#define data_length             		pr_state_ptr->sv_data_length
#define RecvPkptr               		pr_state_ptr->sv_RecvPkptr
#define NoOfLinks               		pr_state_ptr->sv_NoOfLinks
#define HostIPAddress           		pr_state_ptr->sv_HostIPAddress
#define LocalAddress            		pr_state_ptr->sv_LocalAddress
#define rsvpSession             		pr_state_ptr->sv_rsvpSession
#define rsvpSender              		pr_state_ptr->sv_rsvpSender
#define rsvpReserve             		pr_state_ptr->sv_rsvpReserve
#define PSBList                 		pr_state_ptr->sv_PSBList
#define RSBList                 		pr_state_ptr->sv_RSBList
#define TCSBList                		pr_state_ptr->sv_TCSBList




/* Process model interrupt handling procedure */



void
HostRsvpDeamon ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;


	FSM_ENTER (HostRsvpDeamon)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (Init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "Init")
			{
			/*********************************************************************************************************
			** Process Model : HostRsvpDeamon                                                                       **
			** State         : Init                                                                                 **
			** Authors       : Lava K Lavu (llavu@nac.gmu.edu)                                                      **
			**                                                                                                      **
			** This state calls the HostInitialize function which does all the initialiazations that are needed for **
			** this module. See the comments written for this function in the function block.                       **
			**                                                                                                      **
			*********************************************************************************************************/
			
			HostInitialize();
			}


		/** state (Init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "Init")
			{
			}


		/** state (Init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,HostRsvpDeamon)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_TRANSIT_ONLY ((ARRIVAL), 2, state2_enter_exec, ;, "idle")
		/*---------------------------------------------------------*/



		/** state (Arr) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "Arr")
			{
			/*********************************************************************************************************
			** Process Model : HostRsvpDeamon                                                                       **
			** State         : Arr                                                                                  **
			**                                                                                                      **
			** Transition to this state from idle state occurs when there is a packet  arrival. This daemon doesn't **
			** do any reservations  on the host side. It checks if the packet is from the  application and if it is **
			** then looks what  kind of upcall it is. It calls the  appropriate functions for every upcall. See the **
			** comments for every function in the function block.                                                   **
			**                                                                                                      **
			*********************************************************************************************************/
			
			/*
			** Get the input stream from which the packet has arrived. Also get the packet from the incoming packet.
			*/
			InStrm = op_intrpt_strm();
			RecvPkptr = op_pk_get( InStrm );
			
			/*
			** If the packet is arriving from the application, get the application call type and depending on it call
			** the appropriate functions
			*/
			if( InStrm == APPL_STREAM )
			{
				int RecvCallType;
			
				/*
				** Get the recevied application call type and the application packet pointer.
				*/
				op_pk_nfd_get(RecvPkptr, "ApplCallType", &RecvCallType);
				op_pk_nfd_get( RecvPkptr, "AppCall", &ApplPkptr );
			
				switch( RecvCallType )
				{
					case SESSION_MSG :
						/*
						** If the appliation call is session message, pass the aplication call to the MakeSessionCall 
						** function which makes the session call back to the sender.
						*/
						VL	printf("Received Session call from Appl. \n");
						MakeSessionCall( ApplPkptr );
					break;
			
					case SENDER_MSG  :
						/*
						** If the appliation call is sender message, pass  the application call to the MakeSenderCall 
						** function which sends path message to the IP module.
						*/
						VL	printf("Received Sender call from Appl. \n");
						MakeSenderCall( ApplPkptr );
					break;
			
					case RESERVE_MSG :
						/*
						** If the appliation call is reserve message, pass the aplication call to the MakeReserveCall 
						** function which sends Resv message to the IP module.
						*/
						VL	printf("Received Reserve call from Appl. \n");
						MakeReserveCall( ApplPkptr ); 
					break;
			
					case RELEASE_MSG :
						/*
						** If the appliation call is release message, pass the aplication call to the MakeReleaseCall 
						** function which sends PathTear or ResvTear message (depending on whether the application is
						** sender or receiver) to the IP module.
						*/
						MakeReleaseCall( ApplPkptr ); 
					break;
			
					default : 
						printf("Received wrong type of call at the Host Rsvp daemon at %d.%d\n",
							   LocalAddress[0].NetNum, LocalAddress[0].NodeNum );
					break;
				}
			
				/*
				** Destroy the application packet pointer after the necessary actions are performed.
				*/
				op_pk_destroy( ApplPkptr );
			}
			/*
			** Destroy the received upcall packet pointer after the necessary actions are performed.
			*/
			op_pk_destroy( RecvPkptr );
			}


		/** state (Arr) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "Arr")
			{
			}


		/** state (Arr) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,HostRsvpDeamon)
	}




void
HostRsvpDeamon_svar (prs_ptr,var_name,var_p_ptr)
	HostRsvpDeamon_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (HostRsvpDeamon_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("RecDataPacket" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RecDataPacket);
	if (Vos_String_Equal ("RsvpUpCalls" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RsvpUpCalls);
	if (Vos_String_Equal ("SessionId" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_SessionId);
	if (Vos_String_Equal ("GrpId" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_GrpId);
	if (Vos_String_Equal ("GrpNo" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_GrpNo);
	if (Vos_String_Equal ("LinkNo" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_LinkNo);
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("InStrm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_InStrm);
	if (Vos_String_Equal ("data_length" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_data_length);
	if (Vos_String_Equal ("RecvPkptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RecvPkptr);
	if (Vos_String_Equal ("NoOfLinks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfLinks);
	if (Vos_String_Equal ("HostIPAddress" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_HostIPAddress);
	if (Vos_String_Equal ("LocalAddress" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_LocalAddress);
	if (Vos_String_Equal ("rsvpSession" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpSession);
	if (Vos_String_Equal ("rsvpSender" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpSender);
	if (Vos_String_Equal ("rsvpReserve" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpReserve);
	if (Vos_String_Equal ("PSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_PSBList);
	if (Vos_String_Equal ("RSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RSBList);
	if (Vos_String_Equal ("TCSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TCSBList);

	FOUT;
	}




void
HostRsvpDeamon_diag ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;

	FIN (HostRsvpDeamon_diag ())


	FOUT;
	}




void
HostRsvpDeamon_terminate ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;

	FIN (HostRsvpDeamon_terminate ())


	FOUT;
	}



Compcode
HostRsvpDeamon_init (pr_state_pptr)
	HostRsvpDeamon_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (HostRsvpDeamon_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (HostRsvpDeamon)",
			sizeof (HostRsvpDeamon_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (HostRsvpDeamon_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : HostInitialize                                                                    **
	** Input         : void                                                                              **
	** Output        : void                                                                              **
	** Description   : This  function initializes all the variables, get the  object id of the IP module **
	**				   from  that the IP  address of the host, intializes all the  linked lists that are **
	**                 used for  storing Path State Blocks, Resv State Blocks, and Traffic Control state **
	**                 blocks.                                                                           **
	**                                                                                                   **
	******************************************************************************************************/
	void
	HostInitialize()
	{
		char mess[100];
		int i;
	
		/*
		** Get the object id of this module, object id of the IP module and using this object id get the
		** host's IP address.
		*/
		my_id = op_id_self();
		ip_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE , IP_STREAM );
	    for( i=0; i < 1; ++i )
	    {
	        sprintf( mess,"net_num%1d",i );
	        op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NetNum );
	        sprintf( mess,"node_num%1d",i );
	        op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NodeNum );
			VL	printf("Host LocalAddress %d is %d.%d\n", i, 
						LocalAddress[i].NetNum, LocalAddress[i].NodeNum );
	    }
	
		/*
		** Intialize all the lists
		*/
		ListInit(&PSBList);
		ListInit(&RSBList);
		ListInit(&TCSBList);
	}
	
	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : FillCommonHdr                                                                     **
	** Input         : msgtype(message type), ttl(time to live), address of the common header            **
	** Output        : void                                                                              **
	** Description   : This  function fills the fields of the common header. The fields are message type **
	**				   time-to-live and the RSVP version that is being used.                             **
	**                                                                                                   **
	******************************************************************************************************/
	void
	FillCommonHdr( CommonHdr * ch, int msgtype, int ttl )
	{
		ch->Msg_Type = msgtype;
		ch->Send_TTL = ttl;
		ch->Version = RSVP_CURRENT_VER;
	}
	
	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : FillSession                                                                       **
	** Input         : Session Id, address of the session object                                         **
	** Output        : void                                                                              **
	** Description   : This function fills the fields of the Session  Object. The fields are destination **
	**				   address, protocol Id, destination port                                            **
	**                                                                                                   **
	******************************************************************************************************/
	void
	FillSession( SessionObj *ses, int sid )
	{
		ses->DestAddress = rsvpSession[sid].DestAddress;
		ses->Protocol_Id = rsvpSession[sid].ProtocolId;
		ses->DestPort    = rsvpSession[sid].DestPort;
	}
	
	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : Session                                                                           **
	** Input         : Destination address and port, Protocol Id, Upcall Procedure                       **
	** Output        : Session Id for the session                                                        **
	** Description   : This function assigns a session id for the session. This session id is used in the**
	**				   MakeSessionCall function to send this session id to the application. Also it logs **
	**                 all the information of the session in the rsvpsession directory, so that the rsvp **
	**                 daemon can use this information when it gets the upcalls from the application.    **
	**                                                                                                   **
	******************************************************************************************************/
	int 
	Session( IPAddress da, int pid, int dp, void (*Up)() )
	{
		int i;
		static int firstFlag=0;
	
		if( !firstFlag )
		{
			firstFlag=1;
			memset( rsvpSession, 0, sizeof(RsvpSession)*MAXNOOFSESSIONS );
		}
	
		for(i=1; i<MAXNOOFSESSIONS; ++i)
		{
			if( rsvpSession[i].isfull == 0 )
			{
				rsvpSession[i].isfull = 1;
				rsvpSession[i].DestAddress = da;
				rsvpSession[i].ProtocolId = pid;
				rsvpSession[i].DestPort = dp;
				rsvpSession[i].UpcallProc = Up;
	
				VL HRVL	
					printf("Session Information is, DestAddress = %d.%d, DestPort = %d\n", 
							rsvpSession[i].DestAddress.NetNum, rsvpSession[i].DestAddress.NodeNum,
							rsvpSession[i].DestPort );
	
				return(i);
			}
		}
		return(-1);
	}
	
	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : MakeSessionCall                                                                   **
	** Input         : Packet pointer to the application received by this module                         **
	** Output        : void                                                                              **
	** Description   : This function gets the  destination address, port and protocol id from the packet **
	**				   call the Session funtion using these values. It makes an upcall to the application**
	**				   with the session id returned by the Session function.                             **
	**                                                                                                   **
	******************************************************************************************************/
	void
	MakeSessionCall( Packet* ApplPkptr )
	{
	    int tmpSessionId;
	    int pid, dest_port;
	    int dest_net, dest_node;
		IPAddress da;
		Packet* UpcallPkptr;
	 
		/*
		** Get the destination addr. port from the received application packet
		*/
	    op_pk_nfd_get( ApplPkptr, "DestNet", &dest_net );
	    op_pk_nfd_get( ApplPkptr, "DestNode", &dest_node );
	    op_pk_nfd_get( ApplPkptr, "DestPort", &dest_port );
	 
		da.NetNum = dest_net;
		da.NodeNum = dest_node;
	    tmpSessionId = Session( da, pid, dest_port, NULL );		/*       Call the Session function       */
	 
		/*
		** Create the upcall packet and assign the session id returned by the Session  function to Session 
		** Id field and specify that this packet is Session upcall.
		*/
	    UpcallPkptr = op_pk_create_fmt("UPCALL_SESSION");
	    op_pk_nfd_set( UpcallPkptr, "UpcallType", SESSION_MSG );
	    op_pk_nfd_set( UpcallPkptr, "SessionId", tmpSessionId );
	 
	    op_pk_send( UpcallPkptr, APPL_STREAM );				   /*    Send the upcall to the application */
	
		VL	HRVL
			printf("Session Id that was passed to UPcall is %d\n", tmpSessionId); 
	}
	
	/*******************************************************************************************************
	**                                                                                                    **
	** Function Name : MakeSenderCall                                                                     **
	** Input         : Packet pointer to the application received by this module                          **
	** Output        : void                                                                               **
	** Description   : This function gets all the source information, the ttl, bandwidth and delay values **
	**				   with which the sender wants  to send the data to the session and the session id of **
	**				   the session to which it wants to send the data from the received  application call **
	**				   It logs  the sender  information into a array called rsvpSender which is used when **
	**				   terminate the session(to see if PathTear or ResvTear should be sent). It fills all **
	**				   the fields of the Path message using the information obtained from the application **
	**				   call and sends the path message to the IP module. To send the path message a func. **
	**				   called SendPktToIP is used. This function is implemented in the Router RSVP daemon **
	**				   See the comments in the RsvpOnRouter process model.                                **
	**                                                                                                    **
	*******************************************************************************************************/
	void
	MakeSenderCall( Packet* ApplPkptr )
	{
		int i, sid, src_net, src_node, src_port, send_ttl;
		PathMessage * PMEle;
		IPAddress sa;
		static int firstFlag=0;
		SenderTemplateObj sto;
		SenderTspecObj sts;
		Packet* ipdgramptr;
		Packet* PathMsgptr;
		int data_len;
	
		/*
		** Intialize all the elements that are being used to set the path message fields.
		*/
		PMEle = (PathMessage *) op_prg_mem_alloc( sizeof(PathMessage) );
		memset( PMEle, 0, sizeof(PathMessage) );
		memset( &sto, 0, sizeof(SenderTemplateObj) );
		memset( &sts, 0, sizeof(SenderTspecObj) );
	
		/*
		** Get all the attributes of the sender from the sender call received from the application
		*/
		op_pk_nfd_get( ApplPkptr, "SessionId", &sid );
		op_pk_nfd_get( ApplPkptr, "SourceNet", &src_net );
		op_pk_nfd_get( ApplPkptr, "SourceNode", &src_node );
		op_pk_nfd_get( ApplPkptr, "SourcePort", &src_port );
		op_pk_nfd_get( ApplPkptr, "ttl", &send_ttl );
		op_pk_nfd_get( ApplPkptr, "BandWidth", &sts.req_bandwidth );
		op_pk_nfd_get( ApplPkptr, "Delay", &sts.link_delay );
	
		sa = LocalAddress[0];
		sto.SrcAddress = sa;
		sto.SrcPort = src_port;
	
		VL	HRVL
			printf("Received a Sender call with sid %d\n", sid); 
	
		/*
		** Log the sender information into the rsvpSender directory
		*/
		rsvpSender[sid].SessionID = sid;
	    rsvpSender[sid].SrcAddress = sa;
	    rsvpSender[sid].SrcPort = src_port;
	    op_prg_mem_copy( &sto, &(rsvpSender[sid].Sender_Template_Object), sizeof(SenderTemplateObj) );
	    op_prg_mem_copy( &sts, &(rsvpSender[sid].Sender_Tspec_Object), sizeof(SenderTspecObj) );
	    rsvpSender[sid].TTL = send_ttl;
	
		/* 
		** Create Path Message. Fill in all the fields from the information obtained above.
		*/
		FillCommonHdr( &PMEle->common_header, PATH_MESSAGE, send_ttl );    /*  Fill the Common Header   */
		FillSession( &(PMEle->Session_Object), sid );                      /*  Fill the Session Object  */
		PMEle->Rsvp_Hop_Object.Next_Hop_Address = sa;
		PMEle->Rsvp_Hop_Object.Logical_Interface_Handle = sa;
		PMEle->Time_Values_Object.Refresh_Period = REFRESH_PERIOD;
		PMEle->Sender_Descriptor.Sender_Template_Object = sto;
		PMEle->Sender_Descriptor.Sender_Tspec_Object = sts;
	
		SendPktToIP( (char *) PMEle, 0 , PATH_MESSAGE );       /*   Send the Path message to IP layer  */
	}
	
	/*******************************************************************************************************
	**                                                                                                    **
	** Function Name : MakeReserveCall                                                                    **
	** Input         : Packet pointer to the application received by this module                          **
	** Output        : void                                                                               **
	** Description   : This function gets all the source information, sessoin id of the session, the ttl, **
	**                 bandwidth and delay  values with which the receiver wants to receive the data from **
	**                 the  received application call. It logs the sender information into a array called **
	**                 rsvpReserve which is used when terminating the session (to decide whether PathTear **
	**                 or ResvTear should be sent). It fills all the fields of the Resv message using the **
	**                 information obtained from the aplication call and sends the Resv message to the IP **
	**                 module. To send  the Resv message a function called SendResvPktToIP  is used. This **
	**                 function is implemented in the Router RSVP daemon. See the comments in the process **
	**                 model RsvpOnRouter.                                                                **
	**                                                                                                    **
	*******************************************************************************************************/
	void
	MakeReserveCall( Packet* ApplPkptr)
	{
		int sid;
		static int firstFlag=0;
		ResvMessage * rmEle;
		FlowSpecObj fs;
		FilterSpecObj fls;
	
		/*
		** Intialize all the elements that are being used to set the Resv message fields.
		*/
		rmEle = (ResvMessage *) op_prg_mem_alloc( sizeof(ResvMessage) );
		printf("Allocated memory to rmEle \n");
		memset( rmEle, 0, sizeof(ResvMessage) );
		printf("Memset memory of rmEle \n");
		memset( &fs, 0, sizeof(SenderTspecObj) );
		memset( &fls, 0, sizeof(SenderTemplateObj) );
	
		/*
		** Log the Receiver information into the rsvpReserve directory
		*/
	    if( !firstFlag )
	    {
	        firstFlag=1;
	        memset( rsvpReserve, 0, sizeof(RsvpReserve)*MAXNOOFSESSIONS );
		}
		rsvpReserve[sid].SessionID = sid;
	    op_prg_mem_copy(&fs, &(rsvpReserve[sid].flow_descriptor_list[0].Flowspec_Object), sizeof(FlowSpecObj));
	    op_prg_mem_copy(&fls, &(rsvpReserve[sid].flow_descriptor_list[0].Filter_Spec_Object), 
						sizeof(FilterSpecObj) );
	
	
		/*
		** Get all the attributes of the Receiver from the reserve call received from the application
		*/
		op_pk_nfd_get( ApplPkptr, "SessionId", &sid );
		if( sid < 1 )
	    {
	        return;
	    }
		op_pk_nfd_get( ApplPkptr, "SenderNet", &fls.SrcAddress.NetNum );
		op_pk_nfd_get( ApplPkptr, "SenderNode", &fls.SrcAddress.NodeNum );
		op_pk_nfd_get( ApplPkptr, "SenderPort", &fls.SrcPort );
		op_pk_nfd_get( ApplPkptr, "BandWidth", &fs.req_bandwidth );
		op_pk_nfd_get( ApplPkptr, "Delay", &fs.link_delay );
	
		
		VL HRVL
			printf("Sending the Reserve Message \n"); 
	
		/*
		** Create the Resv message from the information obtained through the reserve call
		*/
		FillCommonHdr( &rmEle->common_header, RESV_MESSAGE, DEFAULT_RESERVE_TTL ); /* Fill common header */
	    FillSession( &rmEle->Session_Object, sid );                 /*         Fill Session Object       */
		rmEle->Rsvp_Hop_Object.Logical_Interface_Handle.NetNum = LocalAddress[0].NetNum;
		rmEle->Rsvp_Hop_Object.Logical_Interface_Handle.NodeNum = GwyAddr;
		rmEle->flow_descriptor_list[0].Flowspec_Object = fs;
		rmEle->flow_descriptor_list[0].Filter_Spec_Object = fls;
	
		/* 
		** Call SendResvPktToIP which encapsulates this Resv packet into IP datagram and send it to the 
		** IP layer 
		*/
		SendResvPktToIP( (char *) rmEle, rmEle->Rsvp_Hop_Object.Logical_Interface_Handle );
	}
	 
	/******************************************************************************************************
	**                                                                                                   **
	** Function Name : FillSession                                                                       **
	** Input         : Packet pointer to the application received by this module                         **
	** Output        : void                                                                              **
	** Description   : This function checks if this application is  registered as a receiver or a sender **
	**				   If it is a sender it sends  the PathTear message, and if it is  receiver it sends **
	**				   it sends a ResvTear message. It get all the information that is needed to be sent **
	**				   from the rsvpSender/rsvpReserve arrays                                            **
	**                                                                                                   **
	******************************************************************************************************/
	void
	MakeReleaseCall( Packet* ApplPkptr )
	{
		int sid;
	
		/*
		** Get the session id of the session which should be termianted
		*/
		op_pk_nfd_get( ApplPkptr, "SessionId", &sid );
	
		/*
		** If this session id is less than 1 an error has occured. Do not do anything.
		*/
		if( sid < 1 )
			return;
	
		/*
		** Check if  the application is registered as a sender. If it is registered as a sender send a 
		** PathTear message to the IP stream
		*/
		if( rsvpSender[sid].SessionID == sid )
		{
			/* Path Tear */
		}
		/*
		** Check if the application is registered as a receiver. If it is then send a ResvTear message 
		** to IP module.
		*/
		else if( rsvpReserve[sid].SessionID == sid )
		{
			/* Resv Tear */
		}
		else
		{
			return;
		}
	}
